from typing import Any, Callable, Iterable, Iterator, Sequence, Tuple

from eth_typing import TypeStr
from eth_utils.toolz import compose, curry
from web3._utils.abi import ABITypedData, abi_sub_tree, strip_abi_type
from web3._utils.formatters import recursive_map
from web3.types import TValue


@curry
def map_abi_data(
    normalizers: Sequence[Callable[[TypeStr, Any], Tuple[TypeStr, Any]]],
    types: Sequence[TypeStr],
    data: Sequence[Any],
) -> Any:
    """
    This function will apply normalizers to your data, in the
    context of the relevant types. Each normalizer is in the format:

    def normalizer(datatype, data):
        # Conditionally modify data
        return (datatype, data)

    Where datatype is a valid ABI type string, like "uint".

    In case of an array, like "bool[2]", normalizer will receive `data`
    as an iterable of typed data, like `[("bool", True), ("bool", False)]`.

    Internals
    ---

    This is accomplished by:

    1. Decorating the data tree with types
    2. Recursively mapping each of the normalizers to the data
    3. Stripping the types back out of the tree
    """
    return get_mapper(normalizers, types)(data)


_mappers = {}


def strip_abi_types(data):
    return recursive_map(strip_abi_type, data)


# web3py builds the pipeline every call, we cache it here instead
def get_mapper(
    normalizers: Tuple[Callable[[TypeStr, Any], Tuple[TypeStr, Any]], ...],
    types: Tuple[TypeStr, ...],
) -> Tuple[Callable, ...]:
    mapper = _mappers.get((normalizers, types))
    if mapper is None:
        pipeline = (
            # 1. Decorating the data tree with types
            # web3.py implementation is `abi_data_tree(types)` but a lambda is faster than a curried func call
            lambda data: map(abi_sub_tree, types, data),
            # 2. Recursively mapping each of the normalizers to the data
            IteratorProxy,
            *map(get_data_tree_map, normalizers),
            # 3. Stripping the types back out of the tree
            strip_abi_types,
            list,
        )
        mapper = _mappers[(normalizers, types)] = compose(*reversed(pipeline))
    return mapper


class IteratorProxy(Iterable[TValue]):
    """Wraps an iterator to return when iterated upon"""

    def __init__(self, iterator: Iterator[TValue]):
        self.__wrapped = iterator

    def __iter__(self):
        try:
            return self.__dict__.pop("_IteratorProxy__wrapped")
        except KeyError as e:
            raise RuntimeError(f"{type(self).__name__} has already been consumed") from e.__cause__


_data_tree_maps = {}


def get_data_tree_map(
    func: Callable[[TypeStr, Any], Tuple[TypeStr, Any]],
) -> Callable[[Any], ABITypedData]:
    f = _data_tree_maps.get(func)
    if f is None:

        def map_to_typed_data(elements: Any) -> ABITypedData:
            if isinstance(elements, ABITypedData) and elements.abi_type is not None:
                return ABITypedData(func(*elements))
            else:
                return elements

        typed_data_func = lambda data: recursive_map(map_to_typed_data, data)

        f = _data_tree_maps[func] = typed_data_func

    return f



from typing import (
    Awaitable,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypedDict,
    Union,
    cast,
    overload,
)

from async_lru import alru_cache
from async_property import async_cached_property
from eth_typing import Address, BlockNumber, ChecksumAddress, HexStr
from evmspec import AnyTransaction, FilterTrace, Transaction, TransactionRLP, TransactionReceipt
from evmspec.data import TransactionHash, UnixTimestamp, _decode_hook
from evmspec.data._main import DecodeHook
from evmspec.structs.block import TinyBlock
from evmspec.structs.log import Log
from evmspec.structs.receipt import Status
from hexbytes import HexBytes
from msgspec import Raw, Struct, ValidationError, json
from web3._utils.blocks import select_method_for_block_identifier
from web3._utils.rpc_abi import RPC
from web3.eth import AsyncEth, BaseEth
from web3.method import default_root_munger
from web3.types import BlockIdentifier, ENS, RPCEndpoint, TxParams, Wei

from dank_mids._web3.method import (
    WEB3_MAJOR_VERSION,
    MethodNoFormat,
    bypass_formatters,
    _block_selectors,
)
from dank_mids.types import T


# These 2 type defs are in recent, but not all, web3.py versions
CallOverrideParams = TypedDict(
    "CallOverrideParams",
    {
        "balance": Optional[Wei],
        "nonce": Optional[int],
        "code": Optional[Union[bytes, HexStr]],
        "state": Optional[Dict[HexStr, HexStr]],
        "stateDiff": Optional[Dict[HexStr, HexStr]],
    },
    total=False,
)

CallOverride = Dict[ChecksumAddress, CallOverrideParams]


class TraceFilterParams(TypedDict, total=False):  # type: ignore [call-arg]
    after: int
    count: int
    fromAddress: Sequence[Union[Address, ChecksumAddress, ENS]]
    fromBlock: BlockIdentifier
    toAddress: Sequence[Union[Address, ChecksumAddress, ENS]]
    toBlock: BlockIdentifier


class DankEth(AsyncEth):
    _block_cache_ttl = 0.1

    # eth_chainId

    @async_cached_property
    async def chain_id(self) -> int:
        return await self._chain_id()

    _chain_id: MethodNoFormat[Callable[[], Awaitable[int]]]

    try:
        _chain_id = MethodNoFormat(RPC.eth_chainId, is_property=True)
    except TypeError as e:  # NOTE: older web3.py versions cant use `is_property` kwarg
        if str(e) != "__init__() got an unexpected keyword argument 'is_property'":
            raise
        _chain_id = MethodNoFormat(RPC.eth_chainId, mungers=None)

    # eth_blockNumber

    @property
    @alru_cache(maxsize=None, ttl=_block_cache_ttl)
    async def block_number(self) -> BlockNumber:  # type: ignore [override]
        return await self.get_block_number()

    async def get_block_number(self) -> BlockNumber:  # type: ignore [override]
        return await self._get_block_number()  # type: ignore [misc]

    _get_block_number: MethodNoFormat[Callable[[], Awaitable[BlockNumber]]]

    try:
        _get_block_number = MethodNoFormat(RPC.eth_blockNumber, is_property=True)
    except TypeError as e:  # NOTE: older web3.py versions cant use `is_property` kwarg
        if str(e) != "__init__() got an unexpected keyword argument 'is_property'":
            raise
        _get_block_number = MethodNoFormat(RPC.eth_blockNumber, mungers=None)

        _call: MethodNoFormat[
            Callable[
                [TxParams, Optional[BlockIdentifier], Optional[CallOverride]], Awaitable[HexBytes]
            ]
        ]
        _call = MethodNoFormat(RPC.eth_call, mungers=[BaseEth.call_munger])

    async def get_block_timestamp(self, block_identifier: int) -> UnixTimestamp:
        """
        Retrieves only the timestamp from a specific block.

        This method skips decoding the rest of the Block response data.

        Args:
            block_identifier: The block number from which to retrieve the timestamp.

        Example:
            >>> print(await dank_mids.eth.get_block_timestamp(12345678))
        """
        try:  # TypeError: 'str' object cannot be interpreted as an integer
            block_identifier = hex(block_identifier)
        finally:
            block_bytes = await self._get_block_raw(block_identifier, False)
            return json.decode(
                block_bytes, type=_Timestamped, dec_hook=UnixTimestamp._decode_hook
            ).timestamp

    @overload
    async def get_transactions(self, block_identifier: Union[int, HexStr]) -> List[Transaction]: ...
    @overload
    async def get_transactions(
        self, block_identifier: Union[int, HexStr], hashes_only: Literal[True]
    ) -> List[TransactionHash]: ...
    @overload
    async def get_transactions(
        self, block_identifier: Union[int, HexStr], hashes_only: Literal[False]
    ) -> List[Transaction]: ...
    async def get_transactions(
        self, block_identifier: Union[int, HexStr], hashes_only: bool = False
    ) -> Union[List[Transaction], List[TransactionHash]]:
        """
        Retrieves only the transactions from a specific block.

        This method skips decoding the rest of the Block response data.

        Args:
            block_identifier: The block number or hash from which to retrieve the transactions.
            hashes_only: If True, only transaction hashes will be returned.

        Example:
            >>> [print(tx.hash) for tx in await dank_mids.eth.get_transactions(12345678)]
        """
        try:  # TypeError: 'str' object cannot be interpreted as an integer
            block_identifier = hex(block_identifier)  # type: ignore [arg-type, assignment]
        finally:
            block_bytes = await self._get_block_raw(block_identifier, not hashes_only)
            return json.decode(block_bytes, type=TinyBlock, dec_hook=_decode_hook).transactions

    async def get_transaction_receipt(
        self,
        *args,
        decode_to: Type[T] = TransactionReceipt,
        decode_hook: DecodeHook[T] = _decode_hook,
        **kwargs,
    ) -> T:
        """
        Fetches the transaction receipt and decodes it into the specified format.

        Args:
            *args: Additional positional arguments.
            decode_to: The class to which the receipt should be decoded.
            decode_hook: Hook function to assist in decoding.
            **kwargs: Additional keyword arguments.
        """
        receipt_bytes = await self._get_transaction_receipt_raw(*args, **kwargs)
        return json.decode(receipt_bytes, type=decode_to, dec_hook=decode_hook)

    async def get_transaction_status(self, transaction_hash: str) -> Status:
        """
        Retrieves the status of a transaction.

        Args:
            transaction_hash: The hash of the transaction to query.

        Returns:
            The status of the transaction.
        """
        tiny_receipt = await self.get_transaction_receipt(
            transaction_hash,
            decode_to=_Statusable,
            decode_hook=lambda enum_cls, data: enum_cls(data),
        )
        return tiny_receipt.status

    async def trace_filter(
        self,
        filter_params: TraceFilterParams,
        decode_to: Type[T] = List[FilterTrace],
        decode_hook: DecodeHook[T] = _decode_hook,
    ) -> T:
        """
        Returns all traces matching a filter. If the decoding to the specified
        type fails, the method logs problematic traces and re-raises the
        exception as a diagnostic aid.

        Args:
            filter_params: The parameters defining the traces to filter.
            decode_to: The class to which traces should be decoded.
            decode_hook: Hook function to assist in decoding.

        Raises:
            ValidationError: If a trace cannot be decoded.
        """
        traces_bytes = await self._trace_filter(filter_params)
        try:
            return json.decode(traces_bytes, type=decode_to, dec_hook=decode_hook)
        except ValidationError:
            if decode_to.__origin__ is not list:  # type: ignore [attr-defined]
                raise

            traces_raw = json.decode(traces_bytes, type=List[Raw])
            traces = []
            trace_cls = decode_to.__args__[0]  # type: ignore [attr-defined]
            for raw in traces_raw:
                try:
                    traces.append(json.decode(raw, type=trace_cls, dec_hook=decode_hook))
                except ValidationError as e:
                    e.args = *e.args, json.decode(raw)
                    raise

    async def trace_transaction(self, transaction_hash: str) -> List[FilterTrace]:
        """
        Returns all traces produced by a transaction.

        Args:
            transaction_hash: The hash of the transaction to trace.

        Example:
            >>> traces = await dank_mids.eth.trace_transaction('0x...')
        """
        return await self._trace_transaction(transaction_hash)

    _get_transaction_raw: MethodNoFormat[Callable[[HexStr], Awaitable[Raw]]] = MethodNoFormat(f"{RPC.eth_getTransactionByHash}_raw", mungers=[default_root_munger])  # type: ignore [arg-type,var-annotated]

    async def get_transaction(self, transaction_hash: HexStr) -> AnyTransaction:
        """
        Retrieves a transaction by its hash and attempts to decode it.

        Args:
            transaction_hash: The hash of the transaction to retrieve.

        Raises:
            ValidationError: If the transaction cannot be decoded into either
                             `Transaction` or `TransactionRLP` format.
        """
        transaction_bytes = await self._get_transaction_raw(transaction_hash)
        try:
            return json.decode(transaction_bytes, type=Transaction, dec_hook=_decode_hook)
        except ValidationError:
            try:
                return json.decode(transaction_bytes, type=TransactionRLP, dec_hook=_decode_hook)
            except ValidationError as e:
                e.args = *e.args, json.decode(transaction_bytes)
                raise

    async def get_logs(
        self,
        *args,
        decode_to: Type[T] = Tuple[Log, ...],  # type: ignore [assignment]
        decode_hook: DecodeHook[T] = _decode_hook,
        **kwargs,
    ) -> T:
        """
        Fetches logs and decodes them into the specified format.

        Args:
            *args: Additional positional arguments.
            decode_to: The class to which logs should be decoded.
            decode_hook: Hook function to assist in decoding.
            **kwargs: Additional keyword arguments.
        """
        return json.decode(
            await self._get_logs_raw(*args, **kwargs),
            type=decode_to,
            dec_hook=decode_hook,
        )

    meth = MethodNoFormat.default(RPC.eth_getTransactionReceipt)  # type: ignore [arg-type, var-annotated]
    if WEB3_MAJOR_VERSION >= 6:
        _transaction_receipt = meth
    else:
        _get_transaction_receipt = meth

    _get_transaction_receipt_raw = MethodNoFormat.default(
        cast(RPCEndpoint, f"{RPC.eth_getTransactionReceipt}_raw")
    )

    _get_block_raw: MethodNoFormat[Callable[..., Awaitable[Raw]]] = MethodNoFormat(
        method_choice_depends_on_args=select_method_for_block_identifier(
            **{k: f"{v}_raw" for k, v in _block_selectors.items()}
        ),
        mungers=[AsyncEth.get_block_munger],
    )
    _trace_filter = MethodNoFormat.default(cast(RPCEndpoint, f"{RPC.trace_filter}_raw"))
    _trace_transaction = MethodNoFormat.default(RPC.trace_transaction)


# TODO: this is super hacky, make it not.
bypass_formatters(DankEth)


class _Statusable(Struct, frozen=True):

    status: Status


class _Timestamped(Struct, frozen=True):  # type: ignore [call-arg]

    timestamp: UnixTimestamp
    """The Unix timestamp for when the block was collated."""



import logging
from typing import Any

from dank_mids import ENVIRONMENT_VARIABLES


class DummyLogger:
    """
    A dummy logger class that provides a no-op implementation of the info
    method of logging.Logger.

    This class is used to replace the actual logger when demo mode is not
    enabled, saving processing time by avoiding unnecessary logging operations.
    """

    def info(self, *args: Any, **kwargs: Any) -> None:
        """
        Does nothing. Mimics the API of logging.Logger.info by accepting
        any arguments.

        Args:
            *args: Positional arguments (ignored).
            **kwargs: Keyword arguments (ignored).
        """


# Choose between a real logger and a dummy logger based on the demo mode setting
demo_logger: logging.Logger = logging.getLogger("dank_mids.demo") if ENVIRONMENT_VARIABLES.DEMO_MODE else DummyLogger()  # type: ignore [attr-defined, assignment]



from asyncio import get_running_loop
from collections import defaultdict
from functools import lru_cache
from logging import getLogger
from time import time
from typing import Any, Callable, DefaultDict, List, Literal, Optional, Set

import eth_retry
from cchecksum import to_checksum_address
from eth_typing import ChecksumAddress
from evmspec.data import ChainId
from multicall.constants import MULTICALL_ADDRESSES
from multicall.multicall import NotSoBrightBatcher
from web3 import Web3
from web3.types import RPCEndpoint, RPCResponse

from dank_mids import ENVIRONMENT_VARIABLES as ENVS
from dank_mids import _debugging
from dank_mids._batch import DankBatch
from dank_mids._demo_mode import demo_logger
from dank_mids._exceptions import DankMidsInternalError
from dank_mids._requests import JSONRPCBatch, Multicall, RPCRequest, eth_call
from dank_mids._uid import UIDGenerator, AlertingRLock
from dank_mids.helpers._codec import decode_raw
from dank_mids.helpers._errors import log_request_type_switch
from dank_mids.helpers._helpers import w3_version_major, _make_hashable, _sync_w3_from_async
from dank_mids.helpers._multicall import MulticallContract, _get_multicall2, _get_multicall3
from dank_mids.helpers._session import post, rate_limit_inactive
from dank_mids.semaphores import _MethodQueues, BlockSemaphore
from dank_mids.types import BlockId, PartialRequest, RawResponse, Request

logger = getLogger(__name__)
# our new logger logs the same stuff plus more
getLogger("web3.RequestManager").disabled = True
getLogger("web3.RequestManager").propagate = False

instances: DefaultDict[ChainId, List["DankMiddlewareController"]] = defaultdict(list)


class DankMiddlewareController:
    """
    Controller for managing Dank Middleware operations.

    This class handles the core functionality of Dank Mids, including request batching,
    call execution, and error handling.

    See Also:
        :class:`dank_mids.semaphores.BlockSemaphore`: Used for managing concurrency of eth_calls made with the controller.
    """

    pending_rpc_calls: JSONRPCBatch
    """A :class:`~JSONRPCBatch` containing all pending rpc requests."""

    _pending_rpc_calls_append: Callable
    """An alias for `controller.pending_rpc_calls.append`, intended to minimize attr lookups."""

    def __init__(self, w3: Web3) -> None:
        """
        Initialize the DankMiddlewareController.

        Args:
            w3: The Web3 instance used to make RPC requests.
        """
        logger.info("Dank Middleware initializing... Strap on your rocket boots...")

        self._loop = get_running_loop()

        self.w3: Web3 = w3
        """The Web3 instance used to make rpc requests."""

        self.max_jsonrpc_batch_size = int(ENVS.MAX_JSONRPC_BATCH_SIZE)  # type: ignore [call-overload]

        self.sync_w3 = _sync_w3_from_async(w3)
        """A sync Web3 instance connected to the same rpc, used to make calls during init."""

        chainid = self.sync_w3.eth.chain_id
        self.chain_id = chainid
        """The chainid for the currently connected rpc."""

        self.client_version = _get_client_version(self.sync_w3)
        """The client version for the currently connected rpc."""

        # NOTE: We need this mutable for node types that require the full jsonrpc spec
        self.request_type = (
            Request if ENVS.USE_FULL_REQUEST or "reth" in self.client_version else PartialRequest
        )
        """The Struct class the controller will use to encode requests."""

        self._request_type_changed_ts: float = 0.0
        """The time at which the request type was automatically updated by dank's internals. Zero if never updated after init."""

        # NOTE: Ganache does not support state override. Neither does Gnosis Chain.
        self.state_override_not_supported: bool = ENVS.GANACHE_FORK or chainid == 100  # type: ignore [assignment]
        """A boolean that indicates whether the connected rpc supports state override functionality."""

        self.endpoint: str = self.w3.provider.endpoint_uri  # type: ignore [attr-defined]
        """The uri for the connected rpc."""

        self._sort_calls: bool = "tenderly" in self.endpoint or "chainstack" in self.endpoint
        """A boolean that indicates whether calls must be sorted by id in order for dank to work with the connected rpc."""

        self._sort_response: bool = "chainstack" in self.endpoint
        """A boolean that indicates whether a jsonrpc batch response must be sorted by id in order for dank to work with the connected rpc."""

        if "tenderly" in self.endpoint and ENVS.MAX_JSONRPC_BATCH_SIZE > 10:  # type: ignore [operator]
            logger.info("max jsonrpc batch size for tenderly is 10, overriding existing max")
            self.set_batch_size_limit(10)

        self._instance = sum(map(len, instances.values()))
        instances[chainid].append(self)  # type: ignore

        self.mc2 = _get_multicall2(chainid)
        self.mc3 = _get_multicall3(chainid)
        if self.mc2 is None and self.mc3 is None:
            raise NotImplementedError(
                "Dank Mids currently does not support this network.\nTo add support, you just need to submit a PR adding the appropriate multicall contract addresses to this file:\nhttps://github.com/banteg/multicall.py/blob/master/multicall/constants.py"
            )

        self.no_multicall: Set[ChecksumAddress] = set()
        """A set of addresses that have issues when called from the multicall contract. Calls to these contracts will not be batched in multicalls."""

        if multicall := MULTICALL_ADDRESSES.get(chainid):
            self.no_multicall.add(to_checksum_address(multicall))
        if self.mc2:
            self.no_multicall.add(self.mc2.address)
        if self.mc3:
            self.no_multicall.add(self.mc3.address)
        self._latest_mc = self.mc3 or self.mc2

        self.eth_call_semaphores = BlockSemaphore(
            ENVS.method_semaphores["eth_call"]._value, name=f"eth_call {self}"
        )
        """Used for managing concurrency of eth_calls."""

        # semaphores soon to be deprecated for smart queue
        self.method_queues = _MethodQueues(self)
        """Queues for different method types."""

        self.batcher: NotSoBrightBatcher = NotSoBrightBatcher()
        """Batcher for RPC calls."""

        self.batcher.step = ENVS.MAX_MULTICALL_SIZE  # type: ignore [attr-defined]

        self.call_uid: UIDGenerator = UIDGenerator()
        """Unique identifier generator for individual calls."""

        self.multicall_uid: UIDGenerator = UIDGenerator()
        """Unique identifier generator for multicall operations."""

        self.request_uid: UIDGenerator = UIDGenerator()
        """Unique identifier generator for RPC requests."""

        self.jsonrpc_batch_uid: UIDGenerator = UIDGenerator()
        self.pools_closed_lock: AlertingRLock = AlertingRLock(name="pools closed")

        self.pending_eth_calls: DefaultDict[BlockId, Multicall] = defaultdict(
            lambda: Multicall(self)
        )
        """A dictionary of pending :class:`~Multicall` objects by block. The Multicalls hold all pending eth_calls."""

        self._pending_eth_calls_pop = self.pending_eth_calls.pop
        """An alias for `controller.pending_eth_calls.pop`, intended to minimize attr lookups."""
        self._pending_eth_calls_copy = self.pending_eth_calls.copy
        """An alias for `controller.pending_eth_calls.copy`, intended to minimize attr lookups."""
        self._pending_eth_calls_clear = self.pending_eth_calls.clear
        """An alias for `controller.pending_eth_calls.clear`, intended to minimize attr lookups."""
        self._pending_eth_calls_values = self.pending_eth_calls.values
        """An alias for `controller.pending_eth_calls.values`, intended to minimize attr lookups."""

        self._start_new_batch()

    def __repr__(self) -> str:
        """
        Returns a string representation of the DankMiddlewareController instance.

        Returns:
            str: A string containing the instance number, chain ID, and endpoint.
        """
        return f"<DankMiddlewareController instance={self._instance} chain={self.chain_id} endpoint={self.endpoint}>"

    async def __call__(self, method: RPCEndpoint, params: Any) -> RPCResponse:
        """
        Asynchronous method to handle RPC calls.

        This method routes different types of RPC calls to appropriate handlers,
        including specialized handling for eth_call and other methods that may use queues.

        Args:
            method: The RPC method to be called.
            params: The parameters for the RPC call.

        Returns:
            The response from the RPC call.
        """

        await rate_limit_inactive(self.endpoint)

        # eth_call go thru a specialized Semaphore and other methods pass thru unblocked
        if method == "eth_call":
            async with self.eth_call_semaphores[params[1]]:
                # create a strong ref to the call that will be held until the caller completes or is cancelled
                _logger_debug(f"making {self.request_type.__name__} {method} with params {params}")
                if params[0]["to"] in self.no_multicall:
                    return await RPCRequest(self, method, params)
                return await eth_call(self, params)

        # some methods go thru a SmartProcessingQueue, we check those next
        queue = self.method_queues[method]
        _logger_debug(f"making {self.request_type.__name__} {method} with params {params}")

        # no queue, we can make the request normally
        if queue is None:
            return await RPCRequest(self, method, params)

        # queue found, queue up the call and await the future
        try:
            # NOTE: is this a strong enough ref?
            return await queue(self, method, params)  # type: ignore [return-type]
        except TypeError as e:
            if "unhashable type" not in str(e):
                raise
        return await queue(self, method, _make_hashable(params))

    @eth_retry.auto_retry
    async def make_request(
        self, method: str, params: List[Any], request_id: Optional[int] = None
    ) -> RawResponse:
        """
        Makes an RPC request to the Ethereum node.

        This method creates an RPC request with the given method and parameters,
        sends it to the node, and returns the raw response.

        Args:
            method: The RPC method to call.
            params: The parameters for the RPC method.
            request_id: An optional request ID. If not provided, a new ID will be generated.

        Returns:
            The raw response from the Ethereum node.

        Raises:
            Exception: If DEBUG environment variable is set, any exception that occurs during the request is logged and re-raised.
        """
        request = self.request_type(
            method=method, params=params, id=request_id or self.call_uid.next
        )
        try:
            return await post(self.endpoint, data=request, loads=decode_raw)
        except Exception as e:
            if ENVS.DEBUG:
                _debugging.failures.record(
                    self.chain_id,
                    e,
                    "eth_call" if method == "eth_call" else "RPCRequest",
                    "unknown",
                    "unknown",
                    request.data,
                )
            raise

    async def execute_batch(self) -> None:
        """
        Execute a batch of pending calls.

        This method collects all pending eth calls and RPC calls, clears the pending queues,
        and executes them as a single batch.
        """
        with self.pools_closed_lock:  # Do we really need this?  # NOTE: yes we do
            multicalls = self._pending_eth_calls_copy()
            self._pending_eth_calls_clear()
            rpc_calls = self.pending_rpc_calls
        self._start_new_batch()
        demo_logger.info("executing dank batch (current cid: %s)", self.call_uid.latest)  # type: ignore
        batch = DankBatch(self, multicalls, rpc_calls)
        await batch
        demo_logger.info("%s done", batch)

    @property
    def queue_is_full(self) -> bool:
        """
        Check if the queue of pending calls is full.

        Returns:
            True if the queue is full, False otherwise.
        """
        with self.pools_closed_lock:
            if ENVS.OPERATION_MODE.infura:  # type: ignore [attr-defined]
                return sum(map(len, self.pending_rpc_calls)) >= self.max_jsonrpc_batch_size
            eth_calls = sum(map(len, self._pending_eth_calls_values()))
            other_calls = sum(map(len, self.pending_rpc_calls))
            return eth_calls + other_calls >= self.batcher.step

    def _check_request_type(self) -> bool:
        """
        If the controller is still using the PartialRequest type, it will change over to the Request type.

        Returns:
            True if the request type was updated, either now or any time in the past 10 minutes. False otherwise.
        """
        if not self._request_type_changed_ts:
            self.request_type = Request
            self._request_type_changed_ts = time()
            log_request_type_switch()
            return True
        return time() - self._request_type_changed_ts <= 600

    def early_start(self) -> None:
        """
        Initiate processing of queued calls when a full batch is available.

        This method combines pending eth_calls and other RPC calls into a single batch,
        clears the pending Ethereum calls queue, and starts processing the combined
        batch. It's used to optimize processing by starting as soon as enough calls
        are queued to form a full batch.
        """
        with self.pools_closed_lock:
            self.pending_rpc_calls.extend(self._pending_eth_calls_values(), skip_check=True)
            self._pending_eth_calls_clear()
            self.pending_rpc_calls.start()

    def reduce_multicall_size(self, num_calls: int) -> None:
        """
        Decrease the size of multicall batches in response to failures.

        This method is called when a multicall operation fails, allowing the system
        to dynamically adjust the batch size to prevent future failures.

        Args:
            num_calls: The number of calls in the failed multicall operation.
        """
        self._reduce_chunk_size(num_calls, "multicall")

    def reduce_batch_size(self, num_calls: int) -> None:
        """
        Decrease the size of JSON-RPC batches in response to failures.

        Similar to :meth:`reduce_multicall_size`, this method is called when a JSON-RPC
        batch operation fails, allowing for dynamic adjustment of batch sizes.

        Args:
            num_calls: The number of calls in the failed batch operation.
        """
        self._reduce_chunk_size(num_calls, "batch")

    def _reduce_chunk_size(self, num_calls: int, chunk_name: Literal["multicall", "batch"]) -> None:
        """
        Internal method to reduce the size of processing chunks.

        This method is used by both reduce_multicall_size and reduce_batch_size
        to implement the actual size reduction logic.

        Args:
            num_calls: The number of calls in the failed chunk.
            chunk_name: The type of chunk to reduce, either "multicall" or "batch".

        Raises:
            DankMidsInternalError: If an invalid chunk_name is provided.
        """
        new_chunk_size = round(num_calls * 0.99) if num_calls >= 100 else num_calls - 1
        if new_chunk_size < 30:
            logger.warning(
                f"your {chunk_name} batch size is really low, did you have some connection issue earlier? You might want to restart your script. {chunk_name} chunk size will not be further lowered."
            )
            return
        # NOTE: We need the 2nd check because one of the other calls in a batch might have already reduced the chunk size
        if chunk_name == "batch":
            self.set_batch_size_limit(new_chunk_size)
            logger.info("The failed batch had %s calls", num_calls)
            return
        elif chunk_name == "multicall":
            self.set_multicall_size_limit(new_chunk_size)
            logger.info("The failed multicall had %s calls", num_calls)
            return
        raise DankMidsInternalError(ValueError(f"chunk name {chunk_name} is invalid"))

    def set_multicall_size_limit(self, new_limit: int) -> None:
        """
        Set a new limit for the multicall size.

        Args:
            new_limit: The new maximum number of calls in a multicall.
        """
        existing_limit = self.batcher.step
        if new_limit < existing_limit:
            self.batcher.step = new_limit
            logger.warning("multicall size limit reduced from %s to %s", existing_limit, new_limit)
        else:
            logger.info(
                "new multicall size limit %s is not lower than existing limit %s",
                new_limit,
                existing_limit,
            )

    def set_batch_size_limit(self, new_limit: int) -> None:
        """
        Set a new limit for the JSON-RPC batch size.

        Args:
            new_limit: The new maximum number of calls in a JSON-RPC batch.
        """
        existing_limit = self.max_jsonrpc_batch_size
        if new_limit < existing_limit:
            self.max_jsonrpc_batch_size = new_limit
            logger.warning(
                "jsonrpc batch size limit reduced from %s to %s", existing_limit, new_limit
            )
        else:
            logger.info(
                "new jsonrpc batch size limit %s is not lower than existing limit %s",
                new_limit,
                existing_limit,
            )

    @lru_cache(maxsize=1024)
    def _select_mcall_target_for_block(self, block) -> MulticallContract:
        """
        Select the appropriate multicall contract for a given block.

        Args:
            block: The block number or 'latest'.

        Returns:
            The selected multicall contract.
        """
        if block == "latest":
            return self._latest_mc  # type: ignore [return-value]
        mc3 = self.mc3
        if mc3 and not mc3.needs_override_code_for_block(block):
            return mc3
        # We don't care if mc2 needs override code, mc2 override code is shorter
        return self.mc2 or mc3  # type: ignore [return-value]

    def _start_new_batch(self) -> None:
        """Creates a new :class:`~JSONRPCBatch` and updates the :meth:`_pending_rpc_calls_append` alias accordingly."""
        with self.pools_closed_lock:  # Do we really need this?  # NOTE: yes we do
            batch = JSONRPCBatch(self)
            self.pending_rpc_calls = batch
            self._pending_rpc_calls_append = batch.append


@eth_retry.auto_retry
def _get_client_version(sync_w3: Web3) -> str:
    return sync_w3.client_version if w3_version_major >= 6 else sync_w3.clientVersion  # type: ignore [attr-defined]


def _logger_debug(msg: str, *args: Any) -> None: ...


_logger_debug = logger.debug



import re
from functools import cached_property
from time import time
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    DefaultDict,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Mapping,
    NewType,
    Optional,
    Set,
    Tuple,
    Type,
    TypedDict,
    TypeVar,
    Union,
    overload,
)

import evmspec
from dictstruct import DictStruct
from eth_typing import ChecksumAddress, HexStr
from eth_utils.toolz import valmap
from evmspec.data import Address, BlockNumber, ChainId, Wei, uint, _decode_hook
from evmspec.structs.block import BaseBlock, Block, MinedBlock, ShanghaiCapellaBlock
from evmspec.structs.log import Log
from hexbytes import HexBytes
from msgspec import UNSET, Raw, ValidationError
from msgspec.json import decode as json_decode
from msgspec.json import encode as json_encode
from web3.datastructures import AttributeDict
from web3.types import RPCEndpoint, RPCResponse

from dank_mids import constants, stats
from dank_mids._exceptions import (
    BadResponse,
    ChainstackRateLimited,
    ExceedsMaxBatchSize,
    ExecutionReverted,
    OutOfGas,
    PayloadTooLarge,
)

if TYPE_CHECKING:
    from dank_mids._requests import Multicall


T = TypeVar("T")


BlockId = NewType("BlockId", str)
"""A type representing the identifier for a specific block in the blockchain."""

BatchId = Union[int, str]
"""A type representing the identifier for a batch of operations, which can be either an integer or a string."""

Multicalls = Dict[BlockId, "Multicall"]
"""
A dictionary mapping BlockId to Multicall objects.

See Also:
    :class:`dank_mids._requests.Multicall`: The Multicall class used in this mapping.
"""

eth_callParams = TypedDict("eth_callParams", {"to": ChecksumAddress, "data": str})
"""
A typed dictionary representing the parameters for an eth_call.

See Also:
    :meth:`web3.eth.Eth.call`: Web3's method to perform a call without creating a transaction.
"""

OverrideParams = TypedDict("OverrideParams", {"code": str})
"""A typed dictionary representing override parameters."""

JsonrpcParams = List[Union[eth_callParams, BlockId, OverrideParams]]
"""A list of parameters for JSON-RPC calls, which should be eth_callParams, BlockId, and OverrideParams, in that order."""


# This type alias was introduced in web3 v5.28.0 but we like loose deps here so we recreate instead of import.
AsyncMiddleware = Callable[[RPCEndpoint, Any], Coroutine[Any, Any, RPCResponse]]
"""A type alias for asynchronous middleware functions."""

_list_of_stuff = List[Union[str, None, dict, list]]
"""A type alias for a list that can contain strings, None, dictionaries, or lists."""

_dict_of_stuff = Dict[str, Union[str, None, _list_of_stuff, Dict[str, Optional[Any]]]]
"""A type alias for a dictionary with string keys and values that can be strings, None, :obj:`_list_of_stuff`, or dictionaries with optional values."""

_nested_dict_of_stuff = Dict[str, Union[str, None, _list_of_stuff, _dict_of_stuff]]
"""A type alias for a nested dictionary structure."""


class PartialRequest(DictStruct, frozen=True, omit_defaults=True, repr_omit_defaults=True):  # type: ignore [call-arg]
    """
    Represents a partial JSON-RPC request.

    While technially part of a request, we can successfully make requests to many nodes without including the `jsonrpc` field.

    This class leaves off the `jsonrpc` field reduce encoding burden and web traffic.

    This works with many but not all nodes.
    """

    method: str
    """The RPC method to be called."""

    id: Union[str, int]
    """The request identifier."""

    params: Optional[list] = None
    """The parameters for the RPC method call."""

    @property
    def data(self) -> bytes:
        return json_encode(self, enc_hook=_encode_hook)


class Request(PartialRequest):
    """
    Represents a complete JSON-RPC request.

    Inherits from PartialRequest and adds the JSON-RPC version.
    """

    jsonrpc: Literal["2.0"] = "2.0"
    """The JSON-RPC version, always set to "2.0"."""

    @property
    def data(self) -> bytes:
        # we have to do some hacky stuff because omit_defaults kwarg on PartialRequest
        # is preventing jsonrpc field from being included in the encoded bytes
        encoded_omit_defaults = json_encode(self, enc_hook=_encode_hook)
        return encoded_omit_defaults[:-1] + b',"jsonrpc":"2.0"}'


_getattribute = object.__getattribute__


class Error(DictStruct, frozen=True, omit_defaults=True, repr_omit_defaults=True):  # type: ignore [call-arg]
    """
    Represents an error in a JSON-RPC response.
    """

    code: int
    """The error code."""

    message: str
    """The error message."""

    data: Optional[Any] = UNSET
    """Additional error data, if any."""

    def to_dict(self) -> Dict[str, Any]:
        d = {"code": self.code, "message": self.message}
        data = _getattribute(self, "data")
        if data is not UNSET:
            d["data"] = data
        return d


# some devving tools that will go away eventually
_dict_responses: Set[str] = set()
_str_responses: Set[str] = set()


_RETURN_TYPES = {
    "eth_call": HexBytes,
    "eth_chainId": ChainId,
    "eth_getCode": HexBytes,
    "eth_getLogs": List[Log],
    "eth_getBalance": Wei,
    "eth_blockNumber": BlockNumber,
    "eth_accounts": List[Address],
    "eth_getBlockByNumber": Block,
    "eth_getTransactionCount": uint,
    "eth_getTransactionByHash": evmspec.Transaction,
    "eth_getTransactionReceipt": evmspec.FullTransactionReceipt,
    "erigon_getHeaderByNumber": evmspec.ErigonBlockHeader,
    "trace_filter": List[evmspec.FilterTrace],
    "trace_transaction": List[evmspec.FilterTrace],
}
"""
A dictionary mapping RPC method names to their expected return types.
Used to enable more efficient decoding and validation of RPC responses.
"""


_CHAINSTACK_429_ERR_MSG = "You've exceeded the RPS limit available on the current plan."


class PartialResponse(DictStruct, frozen=True, omit_defaults=True, repr_omit_defaults=True):
    """
    Represents a partial JSON-RPC response.

    We use these to more efficiently decode responses from the node.
    """

    result: Raw = None  # type: ignore
    """The result of the RPC call, if successful."""

    error: Optional[Error] = None
    """The error object, if the call failed."""

    @cached_property
    def exception(self) -> Exception:
        """If the rpc response contains an 'error' field, returns a specialized exception for the specified rpc error."""
        if self.error is None:
            raise AttributeError(f"{self} did not error.")

        message = self.error.message
        if "out of gas" in message:
            return OutOfGas(self)
        if "execution reverted" in message or message in {
            "invalid opcode: INVALID",
            "EVM error: InvalidFEOpcode",
        }:
            return ExecutionReverted(self)
        if self.payload_too_large:
            return PayloadTooLarge(self)
        if re.search(r"batch limit (\d+) exceeded", message):
            return ExceedsMaxBatchSize(self)
        if message == "invalid argument 1: hex string without 0x prefix":
            return TypeError(
                message,
                "DANKMIDS NOTE: You're probably passing what should be an integer type as a string type. "
                "The usual culprit is a block number.",
            )
        # chainstack doesnt return a response with status code 429 when we reach rate limits, so we need to handle it specifically here instead of in the usual place
        if _CHAINSTACK_429_ERR_MSG in message:
            return ChainstackRateLimited(self)
        return BadResponse(self)

    @property
    def payload_too_large(self) -> bool:
        return any(map(self.error.message.__contains__, constants.TOO_MUCH_DATA_ERRS))

    def to_dict(self, method: Optional[RPCEndpoint] = None) -> RPCResponse:  # type: ignore [override]
        """Returns a complete dictionary representation of this response ``Struct``."""
        data: RPCResponse = {
            key: self.decode_result(method=method, caller=self) if key == "result" else value
            for key, value in self.items()
            if value is not None
        }
        return data

    def decode_result(
        self, method: Optional[RPCEndpoint] = None, *, caller=None
    ) -> Union[HexBytes, Wei, uint, ChainId, BlockNumber, AttributeDict]:
        # NOTE: These must be added to the `_RETURN_TYPES` constant above manually
        if method and (typ := _RETURN_TYPES.get(method)):
            if method in (
                "eth_call",
                "eth_blockNumber",
                "eth_getCode",
                "eth_getBlockByNumber",
                "eth_getTransactionReceipt",
                "eth_getTransactionCount",
                "eth_getBalance",
                "eth_chainId",
                "erigon_getHeaderByNumber",
            ):
                try:
                    return better_decode(
                        self.result, type=typ, dec_hook=_decode_hook, method=method
                    )
                except Exception as e:
                    if typ is not Block:
                        raise

                    if e.args[0] == "Object contains unknown field `totalDifficulty`":
                        try:
                            # NOTE should we do this??
                            # _RETURN_TYPES[method] = MinedBlock
                            return better_decode(
                                self.result, type=MinedBlock, dec_hook=_decode_hook, method=method
                            )
                        except ValidationError as e2:
                            if e2.args[0] != "Object contains unknown field `baseFeePerGas`":
                                raise
                            result = better_decode(
                                self.result, type=BaseBlock, dec_hook=_decode_hook, method=method
                            )
                            _RETURN_TYPES[method] = BaseBlock  # all blocks on base are BaseBlocks
                            return result

                    elif e.args[0] == "Object contains unknown field `withdrawals`":
                        typ = ShanghaiCapellaBlock

                    elif e.args[0] == "Object contains unknown field `difficulty`":
                        # I've only seen this on OP stack so far, not sure what difficulty means for their chain
                        typ = BaseBlock

                    else:
                        raise

                    return better_decode(
                        self.result, type=typ, dec_hook=_decode_hook, method=method
                    )

            start = time()
            try:
                decoded = better_decode(self.result, type=typ, dec_hook=_decode_hook, method=method)
            except (ValidationError, TypeError) as e:
                stats.logger.log_validation_error(self, e)
                raise

            if caller:
                stats.log_duration(f"decoding {type(caller)} {method}", start)
            return decoded

        # We have some semi-smart logic for providing decoder hints even if method not in `_RETURN_TYPES`
        if method:
            try:
                if method in _dict_responses:
                    decoded = AttributeDict(json_decode(self.result, type=_nested_dict_of_stuff))
                    stats.logger.log_types(method, decoded)
                    return decoded
                elif method in _str_responses:
                    # TODO: finish adding methods and get rid of this
                    stats.logger.devhint("Must add `%s: str` to `_RETURN_TYPES`", method)
                    return json_decode(self.result, type=str)
            except (ValidationError, TypeError) as e:
                stats.logger.log_validation_error(method, e)

        # In this case we can provide no hints, let's let the decoder figure it out
        decoded = json_decode(self.result)
        if isinstance(decoded, str):
            if method:
                _str_responses.add(method)
            return decoded
        elif isinstance(decoded, dict):
            if method:
                _dict_responses.add(method)
            return AttributeDict(decoded)
        elif isinstance(decoded, list):
            if method is None:
                return decoded
        raise TypeError(
            f"type {type(decoded)} is not supported.  method: {method}  decoded: {decoded}"
        )


class Response(PartialResponse, omit_defaults=True, repr_omit_defaults=True):  # type: ignore [call-arg]
    """
    Represents a complete JSON-RPC response.

    Inherits from PartialResponse and adds the response identifier and JSON-RPC version.
    """

    id: Optional[Union[str, int]] = None
    """The response identifier, matching the request."""

    jsonrpc: Literal["2.0"] = "2.0"
    """The JSON-RPC version, always set to "2.0"."""


class RawResponse:
    """
    Wraps a Raw object that we know represents a Response with a `decode` helper method.
    A `RawResponse` is a properly shaped response for one rpc call, received back from a jsonrpc batch request.
    They represent either a successful or a failed response, stored as pre-decoded bytes.
    """

    def __init__(self, raw: Raw) -> None:
        self._raw = raw
        """The :class:`Raw` object wrapped by this wrapper."""

    @overload
    def decode(self, partial: Literal[True]) -> PartialResponse: ...
    @overload
    def decode(self, partial: Literal[False] = False) -> Response: ...
    def decode(self, partial: bool = False) -> Union[Response, PartialResponse]:
        """Decode the wrapped :class:`Raw` object into a :class:`Response` or a :class:`PartialResponse`."""
        return better_decode(self._raw, type=PartialResponse if partial else Response)

    __slots__ = ("_raw",)


MulticallChunk = Tuple[ChecksumAddress, HexBytes]

JSONRPCBatchRequest = List[Request]
# NOTE: A PartialResponse result implies a failure response from the rpc.
JSONRPCBatchResponse = Union[List[RawResponse], PartialResponse]
# We need this for proper decoding.
JSONRPCBatchResponseRaw = Union[List[Raw], PartialResponse]


StrEncodable = Union[ChecksumAddress, HexStr]
Encodable = Union[int, StrEncodable, HexBytes, bytes]

RpcThing = Union[HexStr, List[HexStr], Dict[str, HexStr]]


def _encode_hook(obj: Encodable) -> RpcThing:
    """
    A hook function for encoding objects during JSON serialization.

    Args:
        obj: The object to encode.

    Returns:
        The encoded object.

    Raises:
        NotImplementedError: If the object type is not supported for encoding.
    """
    try:
        # We just assume `obj` is an int subclass instead of performing if checks because it usually is.
        return hex(int(obj))  # type: ignore [return-value]
    except TypeError as e:
        # I put this here for AttributeDicts which come from eth_getLogs params
        # but I check for mapping so it can work with user custom classes
        if not isinstance(obj, Mapping):
            raise TypeError(obj, type(obj)) from e
        return valmap(_rudimentary_encode_dict_value, obj)
    except ValueError as e:
        # NOTE: The error is probably this if `obj` is a string:
        # ValueError: invalid literal for int() with base 10:"""
        if not isinstance(obj, HexBytes):
            e.args = *e.args, obj, type(obj)
            raise ValueError(obj, type(obj)) from e
        return obj.hex()  # type: ignore [return-value]


def _rudimentary_encode_dict_value(value):
    # I dont think this needs to be robust, time will tell
    return hex(value) if isinstance(value, int) else value


def better_decode(
    data: Raw,
    *,
    type: Optional[Type[T]] = None,
    dec_hook: Optional[Callable[[Type, object], T]] = None,
    method: Optional[str] = None,
) -> T:
    try:
        return json_decode(data, type=type, dec_hook=dec_hook)
    except (ValidationError, TypeError) as e:
        extra_args = [
            f"type: {type.__module__}.{type.__qualname__}",
            f"result: {json_decode(data)}",
        ]
        if method:
            extra_args.insert(0, f"method: {method}")
        e.args = (*e.args, *extra_args)
        raise



import abc
import logging
import os
from typing import Any, Iterable, Literal

import aiofiles
from a_sync.functools import cached_property_unsafe as cached_property
from aiofiles.base import AiofilesContextManager
from aiofiles.threadpool.text import AsyncTextIOWrapper
from async_lru import alru_cache

from dank_mids.helpers.lru_cache import lru_cache_lite

logger = logging.getLogger("dank_mids.debugging")

Mode = Literal["a"]


class _FileHelper(metaclass=abc.ABCMeta):
    path = f"{os.path.expanduser('~')}/.dank_mids/debug"

    def __init__(self, chainid: int):
        if not isinstance(chainid, int):
            raise TypeError(f"`chainid` must be an integer. You passed {chainid}") from None

        self.chainid = chainid
        """The ID of the blockchain network."""

        self.path += f"/{self.chainid}"
        self.ensure_dir()

    @lru_cache_lite
    def ensure_dir(self) -> None:
        """
        Ensure that the directory for the file exists, creating it if necessary.
        """
        os.makedirs(self.path, exist_ok=True)

    def open(self) -> "AiofilesContextManager[None, None, AsyncTextIOWrapper]":
        """
        Open the file asynchronously.

        Returns:
            An asynchronous context manager for the file.
        """
        logger.info("opening %s with mode %s", self.uri, self.mode)
        return aiofiles.open(self.uri, self.mode)

    @abc.abstractproperty
    def uri(self) -> str:
        """
        Abstract property for the file URI.

        Returns:
            The URI of the file as a string.
        """

    @property
    @abc.abstractmethod
    def mode(self) -> Mode:
        """
        Abstract property for the file open mode.

        Returns:
            The mode to open the file with as a string.
        """


class _CSVWriter(_FileHelper):
    mode = "a"

    @cached_property
    def uri(self) -> str:
        """
        Construct the URI for the CSV file.

        Returns:
            The URI of the CSV file as a string.
        """
        return f"{self.path}/{self.filename}"

    async def write_row(self, *values: Any) -> None:
        """
        Write a row to the CSV file after ensuring headers exist.

        Args:
            *values: The values to write as a row in the CSV file.
        """
        await self._ensure_headers()
        await self._write_row(*values)

    @alru_cache(maxsize=None)
    async def _ensure_headers(self) -> None:
        """
        Ensure that the CSV file has headers, writing them if necessary.

        Writes headers to the CSV file if they do not already exist.
        """
        await self._write_row(*self.column_names, new_line=False)

    async def _write_row(self, *values: Any, new_line: bool = True) -> None:
        """
        Write a row to the CSV file.

        Args:
            *values: The values to write as a row.
            new_line: Whether to start a new line before writing. Defaults to True.
        """
        row = ",".join(str(obj) for obj in values)
        if new_line:
            row = f"\n{row}"
        async with self.open() as file:
            logger.debug("writing row %s to file %s", row, file)
            await file.write(row)

    @abc.abstractproperty
    def filename(self) -> str:
        """
        Abstract property for the CSV file name.

        Returns:
            The filename of the CSV file as a string.
        """

    @abc.abstractproperty
    def column_names(self) -> Iterable[str]:
        """
        Abstract property for the column names in the CSV file.

        Returns:
            An iterable of column names for the CSV file.
        """



import functools
from typing import Any, final

from brownie.convert.utils import build_function_selector, build_function_signature


@final
class FunctionABI:
    """
    A class to hold function ABI information.

    This class uses the lru_cache decorator to ensure only one instance is created
    for each unique set of ABI parameters, optimizing memory usage and performance.
    """

    def __init__(self, **abi: Any):
        """
        Initialize a FunctionABI instance with the given ABI information.

        Args:
            **abi: Keyword arguments representing the ABI of the function.
        """

        self.abi = abi
        """
        The complete ABI (Application Binary Interface) of the function.
        This contains all the information about the function's interface.
        """

        self.input_sig = build_function_signature(abi)
        """
        The input signature of the function.
        This is a string representation of the function's parameters and their types.
        """

        self.signature = build_function_selector(abi)
        """
        The function selector (4-byte signature) of the function.
        This is used in Ethereum transactions to identify which function to call.
        """

    @staticmethod
    @functools.lru_cache(maxsize=None)
    def singleton(**abi: Any) -> "FunctionABI":
        """
        Get a singleton FunctionABI to hold function ABI information.

        This class uses the lru_cache decorator to ensure only one instance is created
        for each unique set of ABI parameters, optimizing memory usage and performance.
        """
        return FunctionABI(**abi)



from typing import Any, Callable, Dict, Iterable, Sequence, Tuple, Union

from eth_typing import TypeStr
from eth_utils import to_dict
from eth_utils.curried import apply_formatter_at_index
from eth_utils.toolz import compose
from web3._utils.method_formatters import (
    ABI_REQUEST_FORMATTERS,
    ERROR_FORMATTERS,
    METHOD_NORMALIZERS,
    NULL_RESULT_FORMATTERS,
    PYTHONIC_REQUEST_FORMATTERS,
    STANDARD_NORMALIZERS,
)
from web3._utils.rpc_abi import RPC_ABIS, apply_abi_formatters_to_dict
from web3.types import Formatters, RPCEndpoint, RPCResponse


from dank_mids._web3.abi import get_mapper


return_as_is = lambda x: x


@to_dict
def abi_request_formatters(
    normalizers: Sequence[Callable[[TypeStr, Any], Tuple[TypeStr, Any]]],
    abis: Dict[RPCEndpoint, Union[list, dict]],
) -> Iterable[Tuple[RPCEndpoint, Callable[..., Any]]]:
    for method, abi_types in abis.items():
        if isinstance(abi_types, list):
            yield method, get_mapper(tuple(normalizers), tuple(abi_types))
        elif isinstance(abi_types, dict):
            single_dict_formatter = apply_abi_formatters_to_dict(normalizers, abi_types)
            yield method, apply_formatter_at_index(single_dict_formatter, 0)
        else:
            raise TypeError(f"ABI definitions must be a list or dictionary, got {abi_types!r}")


ABI_REQUEST_FORMATTERS: Formatters = abi_request_formatters(STANDARD_NORMALIZERS, RPC_ABIS)

REQUEST_FORMATTER_MAPS = (
    ABI_REQUEST_FORMATTERS,
    # METHOD_NORMALIZERS needs to be after ABI_REQUEST_FORMATTERS
    # so that eth_getLogs's apply_formatter_at_index formatter
    # is applied to the whole address
    # rather than on the first byte of the address
    METHOD_NORMALIZERS,
    PYTHONIC_REQUEST_FORMATTERS,
)

_request_formatters: Dict[RPCEndpoint, Callable] = {}


def get_request_formatters(
    method_name: Union[RPCEndpoint, Callable[..., RPCEndpoint]],
) -> Callable[..., Any]:
    formatters = _request_formatters.get(method_name)
    if formatters is None:
        combined = (formatter_map.get(method_name) for formatter_map in REQUEST_FORMATTER_MAPS)
        combined = list(filter(None, combined))
        if not combined:
            formatters = return_as_is
        elif len(combined) == 1:
            formatters = combined[0]
        else:
            # NOTE the web3 implementation uses both pipe and compose which I think is unnecessary
            # even compose by itself adds unnecessary overhead if used for only 1 formatter
            formatters = compose(*combined)
        _request_formatters[method_name] = formatters
    return formatters


SuccessFormatter = Callable[[RPCResponse], Any]
ErrorFormatter = Callable[[RPCResponse], Any]
NullFormatter = Callable[[RPCResponse], Any]

ResponseFormatters = Tuple[SuccessFormatter, ErrorFormatter, NullFormatter]

_response_formatters: Dict[RPCEndpoint, ResponseFormatters] = {}


def _get_response_formatters(method: RPCEndpoint) -> ResponseFormatters:
    formatters = _response_formatters[method] = (
        return_as_is,
        ERROR_FORMATTERS.get(method, return_as_is),
        NULL_RESULT_FORMATTERS.get(method, return_as_is),
    )
    return formatters



from asyncio import Task
from itertools import tee
from logging import getLogger
from typing import TYPE_CHECKING, Any, Awaitable, Generator, TypeVar, Union, final

from a_sync import create_task

from dank_mids._exceptions import DankMidsInternalError
from dank_mids._requests import _Batch, JSONRPCBatch, Multicall
from dank_mids.helpers._errors import log_internal_error
from dank_mids.types import Multicalls, RawResponse

if TYPE_CHECKING:
    from dank_mids.controller import DankMiddlewareController

__T = TypeVar("__T")

MIN_SIZE = 1  # TODO: Play with this
"""The minimum size for a batch operation."""

CHECK = MIN_SIZE - 1
"""A constant used for checking batch sizes."""

logger = getLogger(__name__)


def _create_named_task(awaitable: Awaitable[__T]) -> "Task[__T]":
    return create_task(awaitable, name=f"{type(awaitable).__name__} via DankBatch")


@final
class DankBatch:
    """
    A batch of JSON-RPC batches.

    This class represents a collection of multicalls and RPC calls that can be executed as a batch.
    It is used internally by the :class:`DankMiddlewareController` to manage and execute batches of calls.

    Note:
        This class is considered "pretty much deprecated" and needs refactoring in future versions.

    See Also:
        :class:`dank_mids._requests.Multicall`: The Multicall class used in this batch.
        :class:`dank_mids._requests.RPCRequest`: The RPCRequest class used in this batch.
    """

    _awaited: bool = False
    """A flag indicating whether the batch has been awaited."""

    __slots__ = "controller", "multicalls", "rpc_calls", "__dict__"

    def __init__(
        self,
        controller: "DankMiddlewareController",
        multicalls: Multicalls,
        rpc_calls: JSONRPCBatch,
    ):
        self.controller = controller
        """The controller managing this batch."""

        self.multicalls = multicalls
        """A collection of multicalls to be executed."""

        self.rpc_calls = rpc_calls
        """A list of individual RPC calls or multicalls."""

    def __repr__(self):
        return f"<dank_mids.DankBatch object at {hex(id(self))}>"

    def __await__(self) -> Generator[Any, None, None]:
        """
        Executes all operations in the batch.

        This method starts the processing of all multicalls and individual RPC calls
        that have been added to the batch. It marks the batch as awaited to prevent
        duplicate executions.

        Note:
            This method does not wait for the operations to complete. Use the `await` statement
            on the DankBatch instance to wait for completion and handle results.

        Returns a generator that can be awaited to execute the batch.
        """
        if self._awaited:
            raise RuntimeError("The batch has already been awaited")
        self._awaited = True
        for mcall in self.multicalls.values():
            mcall.start(self, cleanup=False)
        for call in self.rpc_calls:
            call._batch = self
        return self._await().__await__()

    async def _await(self) -> None:
        """
        Internal method to await the completion of all coroutines in the batch.

        This method gathers all coroutines in the batch and awaits their completion,
        logging any exceptions that may occur during execution. It's designed to
        handle both successful completions and potential errors gracefully.

        Raises:
            Exception: If any of the coroutines in the batch raise an exception,
                       it will be re-raised after all coroutines have been processed.
        """
        # tee turns 1 iterator into 2
        batches0, batches1 = tee(self.coroutines, 2)

        tasks = map(_create_named_task, batches0)
        last_failure = None
        for batch, task in zip(batches1, list(tasks)):
            try:
                await task
            except Exception as e:
                # we just collect the exceptions for now, we raise later if applicable
                if not isinstance(e, DankMidsInternalError):
                    log_internal_error(logger, batch, e)
                last_failure = task

        if last_failure is not None:
            # raise the last exception if any
            await last_failure

    @property
    def coroutines(self) -> Generator[Union[_Batch, Awaitable[RawResponse]], None, None]:
        """
        Generator that prepares RPC calls and multicalls for batch processing.

        This method combines multicalls and individual RPC calls into one or more JSON-RPC batches,
        considering the batch size limits. The process involves appending calls to a working batch
        and yielding batches when they reach the defined size or when all calls are included.

        Yields:
            A combination of JSON-RPC batches ready for execution.
        """
        # Combine multicalls into one or more jsonrpc batches

        # Start with our existing batch
        working_batch = self.rpc_calls

        # alias since this code runs in tight loops
        batch_append = working_batch.append

        check_len = min(CHECK, self.controller.batcher.step)
        # Go thru the multicalls and add calls to the batch
        for mcall in self.multicalls.values():
            if len(mcall) >= check_len:
                batch_append(mcall, skip_check=True)
            else:
                # NOTE: If a multicall has less than `check_len` calls, we should
                #       just throw the calls into a jsonrpc batch individually.
                working_batch.extend(mcall, skip_check=True)
            if working_batch.is_full:
                yield working_batch
                working_batch = JSONRPCBatch(self.controller)
                batch_append = working_batch.append

        if working_batch:
            if len(working_batch) == 1:
                call = next(iter(working_batch))
                # since we're not going thru the batch code we'll do this here
                working_batch._done.set()
                yield call if type(call) is Multicall else call.make_request()
            else:
                yield working_batch



from itertools import accumulate, chain
from typing import Any, AnyStr, Callable, Final, Iterable, List, Union, Tuple, TypeVar

import msgspec
import msgspec.json
from eth_abi import decoding, encoding
from eth_abi.abi import default_codec
from eth_abi.encoding import DynamicArrayEncoder, TupleEncoder

from dank_mids import types
from dank_mids.types import JSONRPCBatchResponseRaw, MulticallChunk, PartialResponse


__T = TypeVar("__T")

MulticallEncoder = Callable[[Tuple[bool, Iterable[MulticallChunk]]], bytes]

DecodedMulticall = Tuple[int, int, Tuple[Tuple["Success", bytes], ...]]
MulticallDecoder = Callable[..., DecodedMulticall]


# these compile to c constants
Raw: Final = msgspec.Raw
RawResponse: Final = types.RawResponse
ContextFramesBytesIO: Final = decoding.ContextFramesBytesIO
DecodeError: Final = msgspec.DecodeError

json_encode: Final = msgspec.json.encode
json_decode: Final = msgspec.json.decode
encode_uint_256: Final = encoding.encode_uint_256
_encode_hook: Final = types._encode_hook


def decode_raw(data: AnyStr) -> types.RawResponse:
    """
    Decode json-encoded bytes into a `msgspec.Raw` object.

    Args:
        data: The json-encoded bytes to decode.

    Returns:
        The decoded data wrapped in a RawResponse object.
    """
    try:
        return RawResponse(json_decode(data, type=Raw))
    except DecodeError as e:
        e.args = *e.args, data
        raise


def decode_jsonrpc_batch(data: AnyStr) -> Union[PartialResponse, List[types.RawResponse]]:
    """
    Decode json-encoded bytes into a list of response structs, or a single error response struct if applicable.

    Args:
        data: The json-encoded bytes that represent the batch response data.

    Returns:
        Either a PartialResponse if there's an error, or a list of RawResponse objects.
    """
    decoded = json_decode(data, type=JSONRPCBatchResponseRaw)
    return [RawResponse(d) for d in decoded] if isinstance(decoded, list) else decoded


def encode(obj: Any) -> bytes:
    """
    Encode an object to json-encoded bytes.

    This function uses a custom encoding hook defined in `_encode_hook`.

    Args:
        obj: The object to encode.

    Returns:
        The json-encoded bytes representation of the object.

    See Also:
        :func:`msgspec.json.encode`: The underlying JSON encoding function.
    """
    return json_encode(obj, enc_hook=_encode_hook)  # type: ignore [no-any-return]


# multicall specific stuff

_mcall_encoder: Final[MulticallEncoder] = default_codec._registry.get_encoder(
    "(bool,(address,bytes)[])"
)
_array_encoder: Final[DynamicArrayEncoder] = _mcall_encoder.encoders[-1]  # type: ignore [attr-defined]
_item_encoder: Final[TupleEncoder] = _array_encoder.item_encoder

# We don't need to follow the validation code from eth-abi since we guarantee the input types
_mcall_encoder.validate_value = _array_encoder.validate_value = _item_encoder.validate_value = lambda *_: ...  # type: ignore [attr-defined, method-assign]


def __encode_new(values: Iterable[MulticallChunk]) -> bytes:
    encoded_elements, num_elements = __encode_elements_new(values)
    return encode_uint_256(num_elements) + encoded_elements  # type: ignore [no-any-return]


def __encode_elements_new(values: Iterable[MulticallChunk]) -> Tuple[bytes, int]:
    tail_chunks = [_item_encoder(v) for v in values]
    count = len(tail_chunks)
    head_length = 32 * count
    tail_offsets = chain((0,), accumulate((len(chunk) for chunk in tail_chunks[:-1])))
    head_chunks = (encode_uint_256(head_length + tail_offset) for tail_offset in tail_offsets)
    return b"".join(chain(head_chunks, tail_chunks)), count


_array_encoder.encode = __encode_new  # type: ignore [method-assign]
_array_encoder.encode_elements = __encode_elements_new  # type: ignore [method-assign]


_mcall_decoder: Final[MulticallDecoder] = default_codec._registry.get_decoder(
    "(uint256,uint256,(bool,bytes)[])"
).decode


def mcall_encode(data: Iterable[MulticallChunk]) -> bytes:
    return _mcall_encoder((False, data))


# maybe use this success flag to do something later
Success = bool


def mcall_decode(data: PartialResponse) -> Union[List[bytes], Exception]:
    try:
        decoded = _mcall_decoder(ContextFramesBytesIO(data.decode_result("eth_call")))[2]  # type: ignore [arg-type]
    except Exception as e:
        # NOTE: We need to safely bring any Exceptions back out of the ProcessPool
        e.args = (*e.args, data.decode_result() if isinstance(data, PartialResponse) else data)
        return e
    else:
        return [tup[1] for tup in decoded]



from datetime import datetime
from typing import TYPE_CHECKING, List, Literal, Type, Union

from a_sync import ProcessingQueue
from a_sync.functools import cached_property_unsafe as cached_property

from dank_mids._debugging._base import _CSVWriter
from dank_mids.helpers.lru_cache import lru_cache_lite_nonull

if TYPE_CHECKING:
    from dank_mids.types import PartialRequest, Request


@lru_cache_lite_nonull
class FailedRequestWriter(_CSVWriter):
    """
    A class for logging failed requests to a CSV file.

    This class provides functionality to record details of failed requests,
    including the request type, unique identifier, length, error, and request data.

    Usage:
        writer = FailedRequestWriter(1, Exception)
        await writer.record_failure.put((error, "request_type", "request_uid", 10, request_data))
    """

    column_names = "request_type", "request_uid", "request_length", "error", "request_data"
    """The names of the columns in the CSV file."""

    def __init__(self, chainid: int, failure_type: Type[BaseException]):
        """
        Initialize the FailedRequestWriter.

        Args:
            chainid: The ID of the blockchain network.
            failure_type: The type of exception to record.

        Raises:
            TypeError: If chainid is not an integer or failure_type is not a subclass of BaseException.
        """
        super().__init__(chainid)
        if not issubclass(failure_type, BaseException):
            raise TypeError(f"`failure_type` must be an Exception type. You passed {failure_type}")

        self.failure_type = failure_type
        """The type of exception to record."""

        self.record_failure = ProcessingQueue(
            self._record_failure, num_workers=1, return_data=False
        )
        """
        A ProcessingQueue instance used to record failed requests asynchronously.
        It processes failures using the :meth:`FailedRequestWriter._record_failure` method with a single worker.
        """

    @cached_property
    def filename(self) -> str:
        """Generate a unique filename for the CSV file, then cache it for subsequent use."""
        return f"{int(datetime.now().timestamp())}_{self.failure_type.__name__}s.csv"

    async def _record_failure(
        self,
        e: Exception,
        request_type: str,
        request_uid: Union[str, int],
        request_length: Union[int, Literal["unknown"]],
        request_data: Union[List["Request"], List["PartialRequest"], bytes],
    ):
        """
        Record a failed request to the CSV file.

        Args:
            e: The exception that occurred.
            request_type: The type of the request.
            request_uid: The unique identifier of the request.
            request_length: The length of the request.
            request_data: The request data.

        Raises:
            TypeError: If the exception is not of the specified failure_type.
        """
        if not isinstance(e, self.failure_type):
            raise TypeError(e, self.failure_type)
        await self.write_row(request_type, request_uid, request_length, e, request_data)


def record(
    chainid: int,
    e: Exception,
    request_type: str,
    request_uid: Union[int, str],
    request_length: Union[int, Literal["unknown"]],
    request_data: Union[List["Request"], List["PartialRequest"], bytes],
) -> None:
    """
    Record a failed request using a new FailedRequestWriter instance.

    Args:
        chainid: The ID of the blockchain network.
        e: The exception that occurred.
        request_type: The type of the request.
        request_uid: The unique identifier of the request.
        request_length: The length of the request.
        request_data: The request data.
    """
    FailedRequestWriter(chainid, type(e)).record_failure(
        e, request_type, request_uid, request_length, request_data
    )



from asyncio import (
    AbstractEventLoop,
    Future,
    InvalidStateError,
    Task,
    create_task,
    get_running_loop,
    sleep,
)
from logging import DEBUG, getLogger
from time import time
from typing import TYPE_CHECKING, Any, Generator, Optional, Union
from weakref import ProxyType, proxy

from web3.types import RPCResponse

from dank_mids.helpers._errors import error_logger_debug

if TYPE_CHECKING:
    from dank_mids._requests import _RequestBase, _Response


logger = getLogger("dank_mids.future")

_logger_is_enabled_for = logger.isEnabledFor
_logger_log = logger._log

_future_init = Future.__init__
_future_await = Future.__await__
_future_set_result = Future.set_result
_future_set_exc = Future.set_exception


class DebuggableFuture(Future):
    # default values
    _debug_logs_enabled: bool = False
    __debug_daemon_task: Optional["Task[None]"] = None

    # type hints
    _result: Optional[RPCResponse]

    def __init__(self, owner: "_RequestBase", loop: AbstractEventLoop) -> None:
        _future_init(self, loop=loop)
        if _logger_is_enabled_for(DEBUG):
            self._debug_logs_enabled = True
            self._owner: ProxyType["_RequestBase[_Response]"] = proxy(owner)

    def __await__(self) -> Generator[Any, None, RPCResponse]:
        if self._debug_logs_enabled and self.__debug_daemon_task is None:
            self.__debug_daemon_task = create_task(
                coro=self.__debug_daemon(),
                name="DebuggableFuture debug daemon",
            )
        return _future_await(self)

    def set_result(self, value: RPCResponse) -> None:
        # sourcery skip: merge-duplicate-blocks, remove-redundant-if
        if self._loop is get_running_loop():
            try:
                _future_set_result(self, value)
            except InvalidStateError as e:
                if self._result is not None:
                    # We already have a result. We don't really care where it came from, this is fine.
                    return
                elif self._exception is not None:
                    # This can happen if 2 copies of a call are going at once, and one gets a revert as a RawResponse
                    # from a JSONRPC batch and the other gets a ContractLogicError from the single call handler
                    return
                elif self._state == "CANCELLED":
                    error_logger_debug("cannot set result: %s is cancelled", self)
                    return
                else:
                    raise NotImplementedError(f"{self._state} is not a valid state") from e

        # The rest of this code just makes it threadsafe(ish) based on an old idea that never was fully implemented
        # One day I'll fully commit to either finishing it up or ripping it out. For now it stays.
        elif self._state == "PENDING":
            self._loop.call_soon_threadsafe(_future_set_result, self, value)
        elif self._result is not None:
            # We already have a result. We don't really care where it came from, this is fine.
            return
        elif self._exception is not None:
            # This can happen if 2 copies of a call are going at once, and one gets a revert as a RawResponse
            # from a JSONRPC batch and the other gets a ContractLogicError from the single call handler
            return
        elif self._state == "CANCELLED":
            error_logger_debug("cannot set result: %s is cancelled", self)
            return
        else:
            raise NotImplementedError(f"{self._state} is not a valid state")

    def set_exception(self, exc: Union[type, BaseException]) -> None:
        if self._loop is get_running_loop():
            try:
                _future_set_exc(self, exc)
            except InvalidStateError as e:
                if self._result is not None or self._exception is not None:
                    # its kinda odd that we get here at all but who cares, the fut has a result/exception!
                    return
                elif self._state == "CANCELLED":
                    error_logger_debug("cannot set result: %s is cancelled", self)
                    return
                raise

        # The rest of this code just makes it threadsafe(ish) based on an old idea that never was fully implemented
        # One day I'll fully commit to either finishing it up or ripping it out. For now it stays.
        elif self._state == "PENDING":
            self._loop.call_soon_threadsafe(_future_set_exc, self, exc)
        elif self._result is not None or self._exception is not None:
            # its kinda odd that we get here at all but who cares, the fut has a result/exception!
            return
        elif self._state == "CANCELLED":
            error_logger_debug("cannot set result: %s is cancelled", self)
            return
        else:
            raise

    async def __debug_daemon(self) -> None:
        start = time()
        done = self.done
        while not done():
            await sleep(60)
            if not done():
                _logger_log(
                    DEBUG, "%s has not received data after %ss", (self._owner, int(time() - start))
                )



from dank_mids.helpers._gather import gatherish
from dank_mids.helpers._helpers import await_all, setup_dank_w3, setup_dank_w3_from_sync
from dank_mids.helpers.future import DebuggableFuture
from dank_mids.helpers.lru_cache import lru_cache_lite, lru_cache_lite_nonull

__all__ = ["setup_dank_w3", "setup_dank_w3_from_sync"]



from typed_envs import EnvVarFactory

_factory = EnvVarFactory("DANKMIDS")
create_env = _factory.create_env

# This only applies to the oldest of dank's envs
_deprecated_format = EnvVarFactory("DANK_MIDS")



# mypy: disable-error-code="attr-defined, dict-item"
from typing import Dict

import multicall
from eth_typing import BlockNumber
from multicall.constants import Network

TOO_MUCH_DATA_ERRS = (
    "payload too large",
    "content length too large",
    "request entity too large",
    "batch limit exceeded",
)
"""A list of error messages indicating that the request sent to the RPC was too large and must be split up.

These error messages are used to identify when a request needs to be broken into smaller chunks.
"""

RETRY_ERRS = (
    "connection reset by peer",
    "server disconnected",
    "execution aborted (timeout =",
    "batch limit exceeded",
    "request timed out",
)
"""A list of error messages that are expected during normal use and are not indicative of any problem(s).

These errors will be automatically retried until success is achieved.
"""

GAS_LIMIT = multicall.constants.GAS_LIMIT
"""The gas limit constant imported from the :mod:`multicall` library.

This value is used as the default gas limit for multicall operations.
"""

MULTICALL2_OVERRIDE_CODE = multicall.constants.MULTICALL2_BYTECODE
"""The bytecode for the Multicall2 contract.

This is used for state override on blocks before the Multicall2 contract was deployed.
"""

try:
    MULTICALL3_OVERRIDE_CODE = multicall.constants.MULTICALL3_BYTECODE
    """The bytecode for the Multicall3 contract, if it is available in the multicall library.

    If the constant is not available, this will fall back to the Multicall2 bytecode.
    """
except AttributeError:
    MULTICALL3_OVERRIDE_CODE = multicall.constants.MULTICALL2_BYTECODE

MULTICALL2_DEPLOY_BLOCKS: Dict[Network, BlockNumber] = {
    Network.Mainnet: 12336033,
    Network.Fantom: 16572242,
    Network.Arbitrum: 821923,
    Network.Optimism: 722566,
}
"""A dictionary mapping networks to the block numbers where Multicall2 was deployed."""

MULTICALL3_DEPLOY_BLOCKS: Dict[Network, BlockNumber] = {
    Network.Mainnet: 14353601,
    Network.Fantom: 33001987,
    Network.Arbitrum: 7654707,
    Network.Optimism: 4286263,
    Network.Base: 5022,
}
"""A dictionary mapping networks to the block numbers where Multicall3 was deployed."""

# When you get these call responses back from the multicall, we know there was some problem with execution.
# If you make the exact same calls without multicall, you will get an Exception not a response.
# TODO: Delete these
BAD_HEXES = (
    # Chainlink feeds no access
    "0x08c379a0000000000000000000000000000000000000000000000000000000000000002000000000000000000000000000000000000000000000000000000000000000094e6f206163636573730000000000000000000000000000000000000000000000",
    # Mint is paused, ironbank
    "0x08c379a00000000000000000000000000000000000000000000000000000000000000020000000000000000000000000000000000000000000000000000000000000000e6d696e7420697320706175736564000000000000000000000000000000000000",
    # Invalid Ether transfer
    "0x08c379a000000000000000000000000000000000000000000000000000000000000000200000000000000000000000000000000000000000000000000000000000000016496e76616c6964206574686572207472616e7366657200000000000000000000",
    # Non Empty Data
    "0x08c379a00000000000000000000000000000000000000000000000000000000000000020000000000000000000000000000000000000000000000000000000000000000e4e4f4e5f454d5054595f44415441000000000000000000000000000000000000",
    # msg.sig is not assigned to submodule
    "0x08c379a0000000000000000000000000000000000000000000000000000000000000002000000000000000000000000000000000000000000000000000000000000000246d73672e736967206973206e6f742061737369676e656420746f207375626d6f64756c6500000000000000000000000000000000000000000000000000000000",
    # only wrapped native contract could send native token
    "0x08c379a0000000000000000000000000000000000000000000000000000000000000002000000000000000000000000000000000000000000000000000000000000000346f6e6c792077726170706564206e617469766520636f6e747261637420636f756c642073656e64206e617469766520746f6b656e000000000000000000000000",
    # Controller Rejected
    "0x08c379a000000000000000000000000000000000000000000000000000000000000000200000000000000000000000000000000000000000000000000000000000000013434f4e54524f4c4c45525f52454a454354454400000000000000000000000000",
    # Diamond: Function does not exist
    "0x08c379a0000000000000000000000000000000000000000000000000000000000000002000000000000000000000000000000000000000000000000000000000000000204469616d6f6e643a2046756e6374696f6e20646f6573206e6f74206578697374",
    # function should be allowed
    "0x08c379a00000000000000000000000000000000000000000000000000000000000000020000000000000000000000000000000000000000000000000000000000000001a66756e6374696f6e2073686f756c6420626520616c6c6f776564000000000000",
    # Function does not exist
    "0x08c379a00000000000000000000000000000000000000000000000000000000000000020000000000000000000000000000000000000000000000000000000000000001846756e6374696f6e20646f6573206e6f742065786973742e0000000000000000",
)

# Not sure why yet but sometimes a multicall will succeed but one of the results will be a failure for one call that doesn't interrupt the rest of the mcall.
# NOTE: we leave off the '0x' so we can compare raw bytes
# NOTE: The 2nd one here needs to be converted to the first format but I need to encounter one in the wild before I can do that
REVERT_SELECTORS = (b"\x08\xc3y\xa0", b"4e487b71")
"""A list of byte strings representing revert selectors.

These selectors are used to identify specific types of revert errors in Ethereum transactions.
"""



from decimal import Decimal
from typing import TYPE_CHECKING, Literal, Optional, Type, Union

import a_sync
from a_sync.primitives import DummySemaphore, ThreadsafeSemaphore
from a_sync.primitives.locks.prio_semaphore import (
    _AbstractPrioritySemaphore,
    _PrioritySemaphoreContextManager,
)
from eth_typing import HexStr
from web3.types import RPCEndpoint

from dank_mids.helpers.lru_cache import lru_cache_lite, lru_cache_lite_nonull

if TYPE_CHECKING:
    from dank_mids.controller import DankMiddlewareController


class _BlockSemaphoreContextManager(_PrioritySemaphoreContextManager):
    """
    A context manager for block-specific semaphores.

    This class is used internally to manage concurrency for operations
    related to specific blockchain blocks.
    """

    _priority_name = "block"
    """The noun that describes the priority, set to "block"."""

    def __init__(
        self,
        parent: "BlockSemaphore",
        priority: Union[int, float, Decimal],
        name: Optional[str] = None,
    ) -> None:
        if not isinstance(priority, (int, float, Decimal)):
            raise TypeError(priority)
        super().__init__(parent, priority, name)


_TOP_PRIORITY = -1


# NOTE: keep this so we can include in type stubs
# class BlockSemaphore(_AbstractPrioritySemaphore[str, _BlockSemaphoreContextManager]):  # type: ignore [type-var]
class BlockSemaphore(_AbstractPrioritySemaphore):
    """A semaphore for managing concurrency based on block numbers.

    This class extends :class:`_AbstractPrioritySemaphore` to provide block-specific concurrency control.

    Args:
        value: The initial value of the semaphore.
        name: An optional name for the semaphore.

    See Also:
        :class:`_BlockSemaphoreContextManager`: The context manager used by this semaphore.
    """

    _context_manager_class: Type[_BlockSemaphoreContextManager]
    """The context manager class used by this semaphore."""

    _top_priority: Literal[-1]
    """The highest priority value, set to -1."""

    def __init__(self, value=1, *, name=None):
        super().__init__(_BlockSemaphoreContextManager, -1, int(value), name=name)

    def __getitem__(self, block: Union[int, HexStr, Literal["latest", None]]) -> "_BlockSemaphoreContextManager":  # type: ignore [override]
        if isinstance(block, int):
            priority = block
        elif isinstance(block, bytes):
            priority = int(block.hex(), 16)
        elif isinstance(block, str) and "0x" in block:
            priority = int(block, 16)
        elif block not in {None, "latest"}:
            # NOTE: We do this to generate an err if an unsuitable value was provided
            priority = block
        else:
            priority = _TOP_PRIORITY
        return super().__getitem__(priority)


class _MethodQueues:
    """
    A class that manages queues for different RPC methods.

    This class creates and stores SmartProcessingQueues for various RPC methods,
    allowing for controlled processing of method calls.
    """

    def __init__(self, controller: "DankMiddlewareController") -> None:
        from dank_mids import ENVIRONMENT_VARIABLES
        from dank_mids._requests import RPCRequest

        self.controller = controller
        """
        A reference to the DankMiddlewareController instance that this _MethodQueues is associated with.
        """

        self.method_queues = {
            method: a_sync.SmartProcessingQueue(
                RPCRequest, num_workers=sem._value, name=f"{method} {controller}"
            )
            for method, sem in ENVIRONMENT_VARIABLES.method_semaphores.items()
            if method != "eth_call"
        }
        """
        A dictionary mapping RPC method names to their corresponding SmartProcessingQueues.
        These queues are used to manage and process requests for different RPC methods.
        """

        self.keys = self.method_queues.keys()
        """
        A view of the keys (RPC method names) in the method_queues dictionary.
        This allows for efficient iteration over the available method names.
        """

    @lru_cache_lite
    def __getitem__(self, method: RPCEndpoint) -> Optional[a_sync.SmartProcessingQueue]:
        """
        Retrieves the queue for a given RPC method.

        Args:
            method: The RPC method to get the queue for.

        Returns:
            The queue for the method, or None if no specific queue is found.
        """
        return next(
            (self.method_queues[key] for key in self.keys if key in method),
            None,
        )



from typing import Any, Dict, Iterator, List, Literal, NewType, Optional, Tuple, Union, overload

import brownie
from brownie.network.contract import (
    ContractCall,
    ContractTx,
    OverloadedMethod,
    build_function_signature,
)
from brownie.typing import AccountsType
from eth_retry import auto_retry

from dank_mids.brownie_patch.call import _patch_call
from dank_mids.brownie_patch.overloaded import _patch_overloaded_method
from dank_mids.brownie_patch.types import (
    ContractMethod,
    DankContractMethod,
    DankOverloadedMethod,
    _get_method_object,
)
from dank_mids.helpers._helpers import DankWeb3


EventName = NewType("EventName", str)
"""A type representing the name of an event in a smart contract.

See Also:
    :class:`brownie.network.contract.ContractEvents`: Brownie's implementation of contract events.
"""

LogTopic = NewType("LogTopic", str)
"""A type representing a log topic in Ethereum transactions.

See Also:
    :meth:`Web3.eth.get_logs`: Web3.py method for retrieving logs.
"""

Method = NewType("Method", str)
"""A type representing the name of a method in a smart contract.

See Also:
    :class:`brownie.network.contract.ContractMethod`: Brownie's implementation of contract methods.
"""

Signature = NewType("Signature", str)
"""A type representing the signature of a method in a smart contract."""


class Contract(brownie.Contract):
    """
    An extended version of brownie.Contract with additional functionality for Dank Mids.

    This class provides lazy initialization of contract methods and supports
    asynchronous operations through Dank Mids middleware.
    """

    @classmethod
    @auto_retry
    def from_abi(
        cls,
        name: str,
        address: str,
        abi: List[dict],
        owner: Optional[AccountsType] = None,
        persist: bool = True,
    ) -> "Contract":
        """
        Create a new Contract instance from an ABI.

        This method allows for the creation of a Contract instance using a provided ABI,
        which is useful when working with contracts that are not in the project's build
        files and not verified on a block explorer.

        Args:
            name: The name of the contract.
            address: The address of the deployed contract.
            abi: The ABI (Application Binary Interface) of the contract.
            owner: The account that owns this contract instance.
            persist: Whether to persist the contract data to brownie's local db for future use.

        Returns:
            A new Contract instance.
        """
        persisted = brownie.Contract.from_abi(name, address, abi, owner, _check_persist(persist))
        return Contract(persisted.address)

    @classmethod
    @auto_retry
    def from_ethpm(
        cls,
        name: str,
        manifest_uri: str,
        address: Optional[str] = None,
        owner: Optional[AccountsType] = None,
        persist: bool = True,
    ) -> "Contract":
        """
        Create a new Contract instance from an ethPM manifest.

        This method allows for the creation of a Contract instance using an ethPM manifest,
        which is a standardized format for Ethereum smart contract packages.

        Args:
            name: The name of the contract.
            manifest_uri: The URI of the ethPM manifest.
            address: The address of the deployed contract (optional).
            owner: The account that owns this contract instance.
            persist: Whether to persist the contract data to brownie's local db for future use.

        Returns:
            A new Contract instance.
        """
        persisted = brownie.Contract.from_ethpm(
            name, manifest_uri, address, owner, _check_persist(persist)
        )
        return Contract(persisted.address)

    @classmethod
    @auto_retry
    def from_explorer(
        cls,
        address: str,
        as_proxy_for: Optional[str] = None,
        owner: Optional[AccountsType] = None,
        silent: bool = False,
        persist: bool = True,
    ) -> "Contract":
        """
        Create a new Contract instance by fetching the ABI from a block explorer.

        This method is useful for interacting with contracts that are not part of the current project,
        as it automatically fetches the contract's ABI from a block explorer.

        Args:
            address: The address of the deployed contract.
            as_proxy_for: The address of the implementation contract if this is a proxy contract.
            owner: The account that owns this contract instance.
            silent: Whether to suppress console output during the process.
            persist: Whether to persist the contract data to brownie's db for future use.

        Returns:
            A new Contract instance.
        """
        persisted = brownie.Contract.from_explorer(
            address, as_proxy_for, owner, silent, _check_persist(persist)
        )
        return Contract(persisted.address)

    topics: Dict[str, str]
    """A dictionary mapping event names to their corresponding topics."""

    signatures: Dict[Method, Signature]
    """A dictionary mapping method names to their corresponding signatures."""

    def __init__(self, *args, **kwargs):
        """
        Initialize the Contract instance.

        This method sets up lazy initialization for contract methods.
        """
        super().__init__(*args, **kwargs)
        self.__post_init__()

    def __post_init__(self) -> None:
        """
        Get rid of the contract call objects so we can materialize them on a JIT basis.

        This method sets up lazy initialization for contract methods.
        """
        for name in self.__method_names__:
            # these are properties defined on _ContractBase and cannot be written to
            if name not in {"_name", "_owner"}:
                object.__setattr__(self, name, _ContractMethodPlaceholder)

    def __getattribute__(self, name: str) -> DankContractMethod:
        """
        Get a contract method attribute.

        This method implements lazy initialization of contract methods.
        If a method object does not yet exist, it is created and cached.

        Args:
            name: The name of the attribute to get.

        Returns:
            The contract method object.
        """
        try:
            attr = object.__getattribute__(self, name)
        except AttributeError as e:
            raise AttributeError(
                f"Contract '{self._name}' object has no attribute '{name}'"
            ) from e.__cause__
        if attr is _ContractMethodPlaceholder:
            attr = self.__get_method_object__(name)
            object.__setattr__(self, name, attr)
        return attr

    @property
    def __method_names__(self) -> Iterator[str]:
        """List of method names defined in the contract ABI."""
        yield from (i["name"] for i in filter(_is_function_abi, self.abi))

    def __get_method_object__(self, name: str) -> DankContractMethod:
        """
        Get a method object for the given method name.

        This method handles both regular and overloaded contract methods,
        returning an appropriate DankContractMethod object.

        Args:
            name: The name of the method to get.

        Returns:
            The initialized contract method object.
        """
        from dank_mids import web3

        overloaded = list(self.__method_names__).count(name) > 1

        for abi in filter(_is_function_abi, self.abi):
            if abi["name"] != name:
                continue

            full_name = f"{self._name}.{name}"
            sig = build_function_signature(abi)

            natspec: Dict[str, Any] = {}
            if self._build.get("natspec"):
                natspec = self._build["natspec"]["methods"].get(sig, {})

            if overloaded is False:
                return _get_method_object(self.address, abi, full_name, self._owner, natspec)

            # special logic to handle function overloading
            elif overloaded is True:
                overloaded = DankOverloadedMethod(self.address, full_name, self._owner)  # type: ignore [call-arg]

            overloaded._add_fn(abi, natspec)

        return overloaded  # type: ignore [return-value]


def _is_function_abi(abi: dict):
    return abi["type"] == "function"


@overload
def patch_contract(contract: Contract, w3: Optional[DankWeb3] = None) -> Contract: ...
@overload
def patch_contract(
    contract: Union[brownie.Contract, str], w3: Optional[DankWeb3] = None
) -> brownie.Contract: ...
def patch_contract(
    contract: Union[Contract, brownie.Contract, str], w3: Optional[DankWeb3] = None
) -> Union[Contract, brownie.Contract]:
    """
    Patch a contract with async and call batching functionalities.

    Args:
        contract: The contract to patch.
        w3: Optional DankWeb3 instance.

    Returns:
        The patched contract.
    """
    if not isinstance(contract, brownie.Contract):
        contract = brownie.Contract(contract)
    if w3 is None and brownie.network.is_connected():
        from dank_mids import dank_web3 as w3
    if w3 is None:
        raise RuntimeError(
            "You must make sure either brownie is connected or you pass in a Web3 instance."
        )
    for k, v in contract.__dict__.items():
        _patch_if_method(v, w3)
    return contract


def _patch_if_method(method: ContractMethod, w3: DankWeb3) -> None:
    """
    Patch a contract method if it's a ContractCall, ContractTx, or OverloadedMethod.

    Args:
        method: The contract method to patch.
        w3: The DankWeb3 instance to use for async calls.
    """
    if isinstance(method, (ContractCall, ContractTx)):
        _patch_call(method, w3)
    elif isinstance(method, OverloadedMethod):
        _patch_overloaded_method(method, w3)


class _ContractMethodPlaceholder:
    """
    A sentinel object that indicates a Contract has a member by a specific name.

    This class is used internally to represent methods that exist on a contract
    but haven't been fully initialized yet, allowing for lazy loading of method objects.
    """


def _check_persist(persist: bool) -> Literal[True]:
    """
    Check if persistence is enabled and raise an error if it's not.

    This function is used to ensure that contract data persistence is enabled.

    Args:
        persist: Boolean indicating whether to persist a :class:`~Contract` to brownie's local db.

    Returns:
        True if persist is True.

    Raises:
        NotImplementedError: If persist is False, indicating that dank_mids
                             requires persistence to be enabled.
    """
    if not persist:
        raise NotImplementedError("persist: False")
    return persist
